package com.example.p2pchatapplication.activity

import android.content.ComponentName
import android.content.Context
import android.content.Intent
import android.content.ServiceConnection
import android.os.Bundle
import android.os.IBinder
import android.view.View
import android.widget.Toast
import androidx.activity.viewModels
import androidx.appcompat.app.AppCompatActivity
import com.example.p2pchatapplication.adapters.ChatAdapter
import com.example.p2pchatapplication.database.ChatDatabase
import com.example.p2pchatapplication.database.MyDataManager
import com.example.p2pchatapplication.databinding.ActivityChatBinding
import com.example.p2pchatapplication.models.ChatMessage
import com.example.p2pchatapplication.models.SocketClientMessage
import com.example.p2pchatapplication.models.User
import com.example.p2pchatapplication.network.ChatSocketManager
import com.example.p2pchatapplication.network.SocketClient
import com.example.p2pchatapplication.network.SocketManager
import com.example.p2pchatapplication.services.MainService
import com.example.p2pchatapplication.utils.Constants
import com.example.p2pchatapplication.utils.PreferenceManager
import com.example.p2pchatapplication.viewModel.ChatViewModel
import kotlinx.coroutines.GlobalScope
import kotlinx.coroutines.launch
import org.java_websocket.WebSocket
import org.java_websocket.client.WebSocketClient
import org.java_websocket.handshake.ServerHandshake
import org.json.JSONObject
import java.net.URI
import java.util.Date


class ChatActivity : AppCompatActivity() {

    private lateinit var binding: ActivityChatBinding
    private lateinit var chatAdapter: ChatAdapter
    private lateinit var socketClient: WebSocketClient
    private lateinit var serverSocketClient: WebSocketClient
    private lateinit var preferenceManager: PreferenceManager
    private lateinit var chatSocketManager: ChatSocketManager

    private val chatViewModel: ChatViewModel by viewModels()

    private lateinit var mainService: MainService
    private var isBound = false

    private lateinit var chatMessages: List<ChatMessage>
    private lateinit var receiverUser: User
    private var socket: WebSocket? = null

    private val connection: ServiceConnection = object : ServiceConnection {
        override fun onServiceConnected(className: ComponentName, service: IBinder) {
            val binder = service as MainService.LocalBinder
            mainService = binder.getService()
            isBound = true
            serverSocketClient = mainService.serverSocketClient
            sendChatRequest()
        }

        override fun onServiceDisconnected(className: ComponentName) {
            isBound = false
        }
    }

    private fun doBindService() {
        bindService(
            Intent(this@ChatActivity, MainService::class.java),
            connection, BIND_AUTO_CREATE
        )
    }

    override fun onStart() {
        super.onStart()
        doBindService()
    }

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        binding = ActivityChatBinding.inflate(layoutInflater)

        preferenceManager = PreferenceManager(applicationContext)
        chatSocketManager = ChatSocketManager.getInstance()
        chatMessages = ArrayList()
        chatAdapter = ChatAdapter(chatMessages, preferenceManager.getString(Constants.KEY_USERNAME)!!)
        setContentView(binding.root)
        setListeners()
        loadConversationDetails()
        loadMessages()

        if (socket == null) {
            println("Socket for this chat does not yet exist...creating new socket")
            setupSocketClient()
        } else {
            println("Socket for this chat already exists :)")
        }

        binding.chatRecyclerView.setAdapter(chatAdapter)
    }

    private fun setupSocketClient() {
        // val uri = URI("ws://" + receiverUser.privateAddress + ":" + receiverUser.privatePort)
        val uri = URI("ws://" + receiverUser.privateAddress + ":" + receiverUser.privatePort)
        println("connecting to server address: " + uri)
        socketClient = object: WebSocketClient(uri) {
            override fun onOpen(handshakedata: ServerHandshake?) {
                runOnUiThread {
                    binding.indicatorOnline.visibility = View.VISIBLE
                    binding.indicatorOffline.visibility = View.GONE
                }
            }

            override fun onMessage(message: String?) {
                if (message == null) {
                    println("no message in socket chat")
                    return
                }
                val jsonMessage = JSONObject(message)
                val topic = jsonMessage.getString("topic")

                if (topic == "chat-message") {
                    println("received chat message")
                    val data: JSONObject = jsonMessage.getJSONObject("data")

                    val chatMessage = ChatMessage()
                    chatMessage.message = data.getString(Constants.KEY_MESSAGE)
                    chatMessage.timestamp = data.getString(Constants.KEY_TIMESTAMP)
                    chatMessage.sender = data.getString(Constants.KEY_SENDER)
                    chatMessage.receiver = data.getString(Constants.KEY_RECEIVER)

                    // insert message into database

                    GlobalScope.launch {
                        println("inserting message to database")
                        ChatDatabase.getDatabase(applicationContext).chatMessageDao().insert(chatMessage)
                    }
                }
            }

            override fun onClose(code: Int, reason: String?, remote: Boolean) {
                runOnUiThread {
                    binding.indicatorOffline.visibility = View.VISIBLE
                    binding.indicatorOnline.visibility = View.GONE
                }
            }

            override fun onError(ex: java.lang.Exception?) {
                println("error in chat socket client: " + ex?.message)
            }

        }
        socketClient.connect()
    }

    private fun sendChatRequest() {
        val userData = JSONObject()
        userData.put("username", preferenceManager.getString(Constants.KEY_USERNAME))
        val msg = SocketClientMessage("chat-request", userData).toJSONString()
        try {
            socketClient.send(msg)
        } catch (e: Exception) {
            serverSocketClient.send(msg)
        }
    }

    private fun sendMessage() {
        // ⚠️ attention: shit code here 💩💩💩
        val chatMessage = JSONObject()
        chatMessage.put(Constants.KEY_MESSAGE, binding.inputMessage.getText().toString())
        chatMessage.put(Constants.KEY_TIMESTAMP, Date())
        chatMessage.put(Constants.KEY_SENDER, preferenceManager.getString(Constants.KEY_USERNAME))
        chatMessage.put(Constants.KEY_RECEIVER, receiverUser.username)

        val msg = SocketClientMessage("chat-message", chatMessage).toJSONString()

        val chatMessageObj = ChatMessage()
        chatMessageObj.message = binding.inputMessage.getText().toString()
        chatMessageObj.timestamp = chatMessage.getString(Constants.KEY_TIMESTAMP)
        chatMessageObj.sender = preferenceManager.getString(Constants.KEY_USERNAME)
        chatMessageObj.receiver = receiverUser.username

        try {
            // If I am the one accepting the connection, use an existing ws to send message
            // If I am the one requesting the connection, create a new ws connection
            if (socket != null) {
                println("Sending message via existing socket")
                socket!!.send(msg)
            } else {
                println("Sending message via new socket client")
                println(msg)
                socketClient.send(msg)
            }

            binding.inputMessage.text = null
            // insert message into database
            chatViewModel.insert(chatMessageObj)
            binding.chatRecyclerView.smoothScrollToPosition(chatMessage.length() - 1)
            binding.chatRecyclerView.visibility = View.VISIBLE
            binding.progressBar.visibility = View.GONE

        } catch (e: Exception) {
            // try sending message via server if p2p fails
            chatMessage.put("discoverySocketId", receiverUser.discoverySocketId)
            val message = SocketClientMessage("forward-message", chatMessage).toJSONString()
            try {
                serverSocketClient.send(message)
                chatViewModel.insert(chatMessageObj)
                binding.inputMessage.text = null
                binding.chatRecyclerView.smoothScrollToPosition(chatMessage.length() - 1)
                binding.chatRecyclerView.visibility = View.VISIBLE
                binding.progressBar.visibility = View.GONE
            } catch (e: Exception) {
                e.printStackTrace()
                showToast("Cannot send message via server")
            }
            e.printStackTrace()
            showToast("Cannot send message via 2p2")
        }
    }

    private fun loadMessages() {
        loading(true)

        chatViewModel.chatMessages.observe(this) { chatMessages ->
            chatMessages?.let {
                if (it.isEmpty()) {
                    binding.textErrorMessage.text = "This is the beginning of an amazing chat..."
                    binding.textErrorMessage.visibility = View.VISIBLE
                    binding.chatRecyclerView.visibility = View.GONE
                } else {
                    binding.textErrorMessage.visibility = View.GONE
                    binding.chatRecyclerView.visibility = View.VISIBLE
                }
                chatAdapter = ChatAdapter(it, preferenceManager.getString(Constants.KEY_USERNAME)!!)
                binding.chatRecyclerView.adapter = chatAdapter
            }
        }
        loading(false)
    }

    private fun setListeners() {
        binding.imageBack.setOnClickListener { goBack() }
        binding.layoutSend.setOnClickListener { sendMessage() }
    }

    private fun goBack() {
        val intent = Intent(applicationContext, ConversationsActivity::class.java)
        startActivity(intent)
    }

    private fun loading(isLoading: Boolean) {
        runOnUiThread {
            if (isLoading) {
                binding.progressBar.visibility = View.VISIBLE
            } else {
                binding.progressBar.visibility = View.INVISIBLE
            }
        }
    }

    private fun loadConversationDetails() {
        receiverUser = intent.getSerializableExtra(Constants.KEY_USER) as User
        socket = chatSocketManager.getUserSocket(receiverUser.username!!)

        binding.textName.text = receiverUser.username
        binding.imageProfile.text = receiverUser.username!![0].uppercaseChar().toString()
    }

    /*private fun parseTimeStamp(date: Date): String {
        return SimpleDateFormat("dd. mm. yyyy - hh:mm", Locale.getDefault()).format(date)
    }*/

    private fun showToast(message: String) {
        Toast.makeText(applicationContext, message, Toast.LENGTH_SHORT).show()
    }

    override fun onDestroy() {
        super.onDestroy()
        if (isBound) {
            unbindService(connection)
            isBound = false
        }
    }
}