package com.example.p2pchatapplication.activity

import android.content.BroadcastReceiver
import android.content.ComponentName
import android.content.Context
import android.content.Intent
import android.content.IntentFilter
import android.content.ServiceConnection
import android.os.Bundle
import android.os.IBinder
import android.view.View
import android.widget.Toast
import androidx.activity.viewModels
import androidx.appcompat.app.AppCompatActivity
import androidx.lifecycle.Observer
import com.example.p2pchatapplication.adapters.ConversationsAdapter
import com.example.p2pchatapplication.databinding.ActivityConversationsBinding
import com.example.p2pchatapplication.listeners.ConversationListener
import com.example.p2pchatapplication.models.Conversation
import com.example.p2pchatapplication.models.SocketClientMessage
import com.example.p2pchatapplication.models.User
import com.example.p2pchatapplication.network.SocketClient
import com.example.p2pchatapplication.network.SocketManager
import com.example.p2pchatapplication.services.MainService
import com.example.p2pchatapplication.utils.Constants
import com.example.p2pchatapplication.utils.PreferenceManager
import com.example.p2pchatapplication.viewModel.ConversationViewModel
import org.java_websocket.client.WebSocketClient
import org.json.JSONObject

class ConversationsActivity : AppCompatActivity(), ConversationListener {
    private lateinit var binding: ActivityConversationsBinding
    private lateinit var conversationsAdapter: ConversationsAdapter
    private lateinit var preferenceManager: PreferenceManager
    private lateinit var broadcastReceiver: BroadcastReceiver
    private lateinit var serverSocketClient: WebSocketClient
    private lateinit var conversations: List<Conversation>

    private lateinit var mainService: MainService
    private var isBound = false

    private val conversationViewModel: ConversationViewModel by viewModels()

    private val connection: ServiceConnection = object : ServiceConnection {
        override fun onServiceConnected(className: ComponentName, service: IBinder) {
            val binder = service as MainService.LocalBinder
            mainService = binder.getService()
            isBound = true
            serverSocketClient = mainService.serverSocketClient
        }

        override fun onServiceDisconnected(className: ComponentName) {
            isBound = false
        }
    }

    private fun doBindService() {
        bindService(
            Intent(this@ConversationsActivity, MainService::class.java),
            connection, BIND_AUTO_CREATE
        )
    }

    override fun onStart() {
        super.onStart()
        doBindService()
    }

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        binding = ActivityConversationsBinding.inflate(layoutInflater)
        preferenceManager = PreferenceManager(applicationContext)
        conversations = ArrayList()
        conversationsAdapter = ConversationsAdapter(conversations, this)
        binding.conversationsRecyclerView.setAdapter(conversationsAdapter)
        setContentView(binding.getRoot())
        setTitle()
        setListeners()
        loadConversations()

        broadcastReceiver = object: BroadcastReceiver() {
            override fun onReceive(context: Context?, intent: Intent?) {
                if (intent?.action == "USER") {
                    val user = intent.getStringExtra("user")
                    val jsonUser = JSONObject(user!!)

                    println(jsonUser.toString())

                    val userObj = User()
                    userObj.username = jsonUser.getString("username")
                    //userObj.publicAddress = jsonUser.getString("publicAddress")
                    userObj.publicPort = jsonUser.getInt("publicPort")
                    if (jsonUser.has("privateAddress")) {
                        userObj.privateAddress = jsonUser.getString("privateAddress")
                    } else {
                        userObj.privateAddress = null
                    }
                    userObj.privatePort = jsonUser.getInt("privatePort")
                    userObj.discoverySocketId = jsonUser.getString("socketId")

                    val intent = Intent(applicationContext, ChatActivity::class.java)
                    intent.putExtra(Constants.KEY_USER, userObj)
                    startActivity(intent)
                }
            }
        }

        val intentFilter = IntentFilter()
        intentFilter.addAction("USER")
        registerReceiver(broadcastReceiver, intentFilter, RECEIVER_EXPORTED)
    }

    private fun loadConversations() {
        loading(true)

        conversationViewModel.conversations.observe(this, Observer { conversations ->
            conversations?.let {
                if (it.isEmpty()) {
                    binding.textErrorMessage.text = "No recent conversations"
                    binding.textErrorMessage.visibility = View.VISIBLE
                    binding.conversationsRecyclerView.visibility = View.GONE
                } else {
                    binding.textErrorMessage.visibility = View.GONE
                    binding.conversationsRecyclerView.visibility = View.VISIBLE
                }
                conversationsAdapter = ConversationsAdapter(it, this)
                binding.conversationsRecyclerView.adapter = conversationsAdapter
            }
        })
        loading(false)
    }

    private fun setTitle() {
        binding.textUsername.text = "Welcome " + preferenceManager.getString(Constants.KEY_USERNAME)
    }

    private fun setListeners() {
        binding.btnNewConversation.setOnClickListener {
            startActivity(Intent(applicationContext, UsersActivity::class.java))
        }

        binding.imageSignOut.setOnClickListener { signOut() }
    }

    private fun showErrorMessage() {
        runOnUiThread {
            binding.textErrorMessage.text = String.format("%s", "No user available")
            binding.textErrorMessage.visibility = View.VISIBLE
        }
    }

    private fun loading(isLoading: Boolean) {
        runOnUiThread {
            if (isLoading) {
                binding.progressBar.visibility = View.VISIBLE
            } else {
                binding.progressBar.visibility = View.INVISIBLE
            }
        }
    }

    private fun showToast(message: String) {
        Toast.makeText(applicationContext, message, Toast.LENGTH_SHORT).show()
    }

    private fun signOut() {
        preferenceManager.clear()
        startActivity(Intent(applicationContext, RegisterActivity::class.java))
        finish()
        showToast("Singed out")
    }

    override fun onDestroy() {
        super.onDestroy()
    }

    override fun onConversationClicked(conversation: Conversation) {
        val data = JSONObject()
        data.put("username", conversation.username)
        val message = SocketClientMessage("user-data", data)
        serverSocketClient.send(message.toJSONString())
    }
}