package com.example.p2pchatapplication.activity

import android.content.BroadcastReceiver
import android.content.ComponentName
import android.content.Context
import android.content.Intent
import android.content.IntentFilter
import android.content.ServiceConnection
import android.os.Bundle
import android.os.IBinder
import android.view.View
import androidx.activity.viewModels
import androidx.appcompat.app.AppCompatActivity
import com.example.p2pchatapplication.adapters.UsersAdapter
import com.example.p2pchatapplication.databinding.ActivityUsersBinding
import com.example.p2pchatapplication.listeners.UserListener
import com.example.p2pchatapplication.models.Conversation
import com.example.p2pchatapplication.models.SocketClientMessage
import com.example.p2pchatapplication.models.User
import com.example.p2pchatapplication.network.SocketClient
import com.example.p2pchatapplication.network.SocketManager
import com.example.p2pchatapplication.services.MainService
import com.example.p2pchatapplication.utils.Constants
import com.example.p2pchatapplication.utils.PreferenceManager
import com.example.p2pchatapplication.viewModel.ConversationViewModel
import org.java_websocket.client.WebSocketClient
import org.json.JSONArray
import org.json.JSONObject

class UsersActivity : AppCompatActivity(), UserListener {

    private lateinit var binding: ActivityUsersBinding
    private lateinit var preferenceManager: PreferenceManager
    private lateinit var broadcastReceiver: BroadcastReceiver
    private lateinit var serverSocketClient: WebSocketClient
    private lateinit var allUsers: MutableList<User>

    private lateinit var mainService: MainService
    private var isBound = false

    private val conversationViewModel: ConversationViewModel by viewModels()

    private val connection: ServiceConnection = object : ServiceConnection {
        override fun onServiceConnected(className: ComponentName, service: IBinder) {
            val binder = service as MainService.LocalBinder
            mainService = binder.getService()
            isBound = true
            serverSocketClient = mainService.serverSocketClient
            getAllUsers()
        }

        override fun onServiceDisconnected(className: ComponentName) {
            isBound = false
        }
    }

    private fun doBindService() {
        bindService(
            Intent(this@UsersActivity, MainService::class.java),
            connection, BIND_AUTO_CREATE
        )
    }

    override fun onStart() {
        super.onStart()
        doBindService()
    }

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        binding = ActivityUsersBinding.inflate(layoutInflater)
        preferenceManager = PreferenceManager(applicationContext)
        allUsers = ArrayList()
        // usersAdapter = UsersAdapter(onlineUsers, this)
        // binding.usersRecyclerView.setAdapter(usersAdapter)
        setContentView(binding.getRoot())
        setListeners()

        broadcastReceiver = object: BroadcastReceiver() {
            override fun onReceive(context: Context?, intent: Intent?) {
                if (intent?.action == "ALL_USERS") {
                    val users = intent.getStringExtra("users")
                    val jsonUsers = JSONArray(users)
                    val userList = ArrayList<User>()
                    for (i in 0 until jsonUsers.length()) {
                        val user = User()
                        user.username = jsonUsers.getJSONObject(i).getString("username")
                        // user.publicAddress = jsonUsers.getJSONObject(i).getString("publicAddress")
                        user.publicPort = jsonUsers.getJSONObject(i).getInt("publicPort")
                        user.discoverySocketId = jsonUsers.getJSONObject(i).getString("socketId")
                        if (jsonUsers.getJSONObject(i).has("privateAddress")) {
                            user.privateAddress =
                                jsonUsers.getJSONObject(i).getString("privateAddress")
                        } else {
                            user.privateAddress = null
                        }

                        user.privatePort = jsonUsers.getJSONObject(i).getInt("privatePort")
                        userList.add(user)
                    }
                    allUsers = userList

                    if (allUsers.size == 0) {
                        showErrorMessage()
                    } else {
                        val usersAdapter = UsersAdapter(allUsers, this@UsersActivity)
                        binding.usersRecyclerView.setAdapter(usersAdapter)
                    }
                }

            }
        }

        val intentFilter = IntentFilter()
        intentFilter.addAction("ALL_USERS")
        registerReceiver(broadcastReceiver, intentFilter, RECEIVER_EXPORTED)
    }


    private fun getAllUsers() {
        loading(true)
        val socketMessage = SocketClientMessage(Constants.MESSAGE_TOPIC_GET_ALL_USERS, JSONObject())
        serverSocketClient.send(socketMessage.toJSONString())
    }

    private fun setListeners() {
        binding.imageBack.setOnClickListener {
            startActivity(Intent(applicationContext, ConversationsActivity::class.java))
        }
    }

    private fun showErrorMessage() {
        runOnUiThread {
            binding.textErrorMessage.text = String.format("%s", "No users available")
            binding.textErrorMessage.visibility = View.VISIBLE
        }
    }

    private fun loading(isLoading: Boolean) {
        runOnUiThread {
            if (isLoading) {
                binding.progressBar.visibility = View.VISIBLE
            } else {
                binding.progressBar.visibility = View.INVISIBLE
            }
        }
    }

    /*private fun showToast(message: String) {
        Toast.makeText(applicationContext, message, Toast.LENGTH_SHORT).show()
    }*/

    override fun onDestroy() {
        super.onDestroy()
    }

    override fun onUserClicked(user: User) {
        val newConversation = Conversation()
        newConversation.username = user.username!!
        conversationViewModel.insert(newConversation)

        val intent = Intent(applicationContext, ChatActivity::class.java)
        intent.putExtra(Constants.KEY_USER, user)
        startActivity(intent)
        // finish()
    }
}