package com.example.p2pchatapplication.services

import android.app.Service
import android.content.Intent
import android.os.Binder
import android.os.IBinder
import com.example.p2pchatapplication.activity.ConversationsActivity
import com.example.p2pchatapplication.database.ChatDatabase
import com.example.p2pchatapplication.database.MyDataManager
import com.example.p2pchatapplication.models.ChatMessage
import com.example.p2pchatapplication.models.Conversation
import com.example.p2pchatapplication.models.User
import com.example.p2pchatapplication.network.ChatSocketManager
import com.example.p2pchatapplication.utils.Constants
import com.example.p2pchatapplication.utils.PreferenceManager
import kotlinx.coroutines.DelicateCoroutinesApi
import kotlinx.coroutines.GlobalScope
import kotlinx.coroutines.launch
import org.java_websocket.WebSocket
import org.java_websocket.client.WebSocketClient
import org.java_websocket.handshake.ClientHandshake
import org.java_websocket.handshake.ServerHandshake
import org.java_websocket.server.WebSocketServer
import org.json.JSONArray
import org.json.JSONObject
import java.io.Serializable
import java.net.InetSocketAddress
import java.net.URI


class MainService : Service() {
    private lateinit var socketServer: WebSocketServer
    lateinit var serverSocketClient: WebSocketClient
    private lateinit var chatSocketManager: ChatSocketManager
    private lateinit var myDataManager: MyDataManager
    private lateinit var preferenceManager: PreferenceManager

    private val binder = LocalBinder()

    inner class LocalBinder : Binder() {
        fun getService(): MainService = this@MainService
    }

    override fun onBind(intent: Intent?): IBinder {
        return binder
    }

    override fun onCreate() {
        super.onCreate()
        chatSocketManager = ChatSocketManager.getInstance()
        myDataManager = MyDataManager.getInstance()
        preferenceManager = PreferenceManager(applicationContext)
        setupSocketServer()
        setupServerSocketClient()
    }

    @OptIn(DelicateCoroutinesApi::class)
    private fun setupSocketServer() {
        socketServer = object : WebSocketServer(InetSocketAddress(Constants.WS_SERVER_PORT)) {
            override fun onOpen(conn: WebSocket?, handshake: ClientHandshake?) {
                println("peer connected")
            }

            override fun onClose(conn: WebSocket?, code: Int, reason: String?, remote: Boolean) {
                println("peer disconnected")
            }

            override fun onError(conn: WebSocket?, ex: Exception?) {
                println("Cannot start websocket server: " + ex?.message)
            }

            override fun onStart() {
                println("socket server listening on port " + socketServer.port)
            }

            override fun onMessage(conn: WebSocket?, message: String?) {
                if (message == null) {
                    println("No message")
                    return
                }
                val jsonMessage = JSONObject(message)
                val topic = jsonMessage.getString("topic")

                if (topic == "chat-request") {
                    println("received chat request")
                    val data: JSONObject = jsonMessage.getJSONObject("data")
                    val username = data.getString("username")

                    val newConversation = Conversation()
                    newConversation.username = username


                    GlobalScope.launch {
                        println("inserting into database...")
                        ChatDatabase.getDatabase(applicationContext).conversationDao().insert(newConversation)
                    }

                    chatSocketManager.add(username, conn)
                }
                else if (topic == "chat-message") {
                    println("received chat message")
                    val data: JSONObject = jsonMessage.getJSONObject("data")

                    val chatMessage = ChatMessage()
                    chatMessage.message = data.getString(Constants.KEY_MESSAGE)
                    chatMessage.timestamp = data.getString(Constants.KEY_TIMESTAMP)
                    chatMessage.sender = data.getString(Constants.KEY_SENDER)
                    chatMessage.receiver = data.getString(Constants.KEY_RECEIVER)

                    // insert message into database

                    GlobalScope.launch {
                        println("inserting message to database")
                        ChatDatabase.getDatabase(applicationContext).chatMessageDao().insert(chatMessage)
                    }
                }
            }
        }
        socketServer.isReuseAddr = true
        socketServer.start()
    }

    @OptIn(DelicateCoroutinesApi::class)
    private fun setupServerSocketClient() {
        serverSocketClient = object : WebSocketClient(URI(Constants.DISCOVERY_SERVER_ADDRESS)) {
            override fun onOpen(handshakedata: ServerHandshake?) {
                println("connected to discovery server")
                val intent = Intent()
                intent.action = "SERVER_CONNECTED"
                sendBroadcast(intent)
            }

            // receive messages forwarded by server
            override fun onMessage(message: String?) {
                println("received message from server: " + message)
                if (message == null) {
                    println("No message")
                    return
                }
                val jsonMessage = JSONObject(message)
                val topic = jsonMessage.getString("topic")

                if (jsonMessage.getString("topic") == "chat-message") {
                    val data: JSONObject = jsonMessage.getJSONObject("data")

                    val chatMessage = ChatMessage()
                    chatMessage.message = data.getString(Constants.KEY_MESSAGE)
                    chatMessage.timestamp = data.getString(Constants.KEY_TIMESTAMP)
                    chatMessage.sender = data.getString(Constants.KEY_SENDER)
                    chatMessage.receiver = data.getString(Constants.KEY_RECEIVER)

                    GlobalScope.launch {
                        ChatDatabase.getDatabase(applicationContext).chatMessageDao().insert(chatMessage)
                    }
                }
                else if (topic == "register-response") {
                    val data: JSONObject = jsonMessage.getJSONObject("data")

                    val code: String = jsonMessage.getString("code")
                    val message: String = data.getString("message")
                    val me = data.getJSONObject("me")

                    val user = User()
                    user.username = me.getString("username")
                    // user.publicAddress = me.getString("publicAddress")
                    user.publicPort = me.getInt("publicPort")
                    user.privatePort = me.getInt("privatePort")
                    user.discoverySocketId = me.getString("socketId")
                    if (me.has("privateAddress")) {
                        user.privateAddress = me.getString("privateAddress")
                    } else {
                        user.privateAddress = null
                    }

                    myDataManager.setData(user)

                    if (code == "success") {
                        preferenceManager.putString(Constants.KEY_USERNAME, user.username)
                        preferenceManager.putBoolean(Constants.KEY_IS_SIGNED_IN, true)

                        // send register success
                        val intent = Intent()
                        intent.action = "REGISTER_SUCCESS"
                        intent.putExtra("message", message)
                        sendBroadcast(intent)
                    } else {
                        val intent = Intent()
                        intent.action = "REGISTER_FAILED"
                        intent.putExtra("message", message)
                        sendBroadcast(intent)
                    }
                }
                else if (topic == "all-users-response") {
                    val users: JSONArray = jsonMessage.getJSONArray("data")
                    val code: String = jsonMessage.getString("code")

                    if (code == "success") {
                        val intent = Intent()
                        intent.action = "ALL_USERS"
                        intent.putExtra("users", users.toString())
                        sendBroadcast(intent)
                    }
                }
                else if (jsonMessage.getString("topic") == "user-data-response") {
                    val user: JSONObject = jsonMessage.getJSONObject("data")
                    val code: String = jsonMessage.getString("code")

                    if (code == "success") {
                        val intent = Intent()
                        intent.action = "USER"
                        intent.putExtra("user", user.toString())
                        sendBroadcast(intent)
                    }
                }


            }
            override fun onClose(code: Int, reason: String?, remote: Boolean) {
                println("connection with discovery server closed")
            }

            override fun onError(ex: java.lang.Exception?) {
                println("Error in serverSocketClient: " + ex?.message)
            }
        }
        serverSocketClient.connect()
    }
}

